/*-
 * Copyright (C)2003..2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2003..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: test.c,v 1.23 2014/09/07 08:25:33 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT 3

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "../mife.h"
#define SEQN(i) ((((u_int32_t)(i) ^ 0x33CC55AA) >> (((u_int32_t)(i) & 3) * 7 + 2) | 0x80) & 0xFF)
#define SEEN(mf, of, i) (((u_char*)mife_point((mf), MIFE_POINOFF, (off_t)(of)))[i] & 0xFF)

int
main(int argc, char **argv) {
    mife_descriptor *mf;
    ssize_t c, bl, pl;
    int i, fd, pr, j;
    BLIN_flag flags;
    mife_tim tt;
    u_char *pos;
    char *eoln;
    off_t of;

    if  (!argv) exit(EX_USAGE);
    flags = MIFE_BUFX;
    fd = fileno(stdin);
    eoln = NULL;
    bl = 65536;
    pr = 0;
    tt = 0;
    while ((i = getopt(argc, argv, "b:fl:pqr:st:vx:")) > 0) {
        switch (i) {
        case 'b': /******** block size         ********/
                  bl = strtoul(optarg, NULL, 0);
                  break;
        case 'f': /******** zero at end        ********/
                  flags |= MIFE_FULL;
                  break;
        case 'l': /******** buffer space       ********/
                  flags |= MIFE_BUFX;
                  flags |= (strtoul(optarg, NULL, 0) << 8) & MIFE_BUFL;
                  break;
        case 'p': /******** print block bound  ********/
                  pr = 1;
                  break;
        case 'q': /******** quiet              ********/
                  flags &= ~BLIN_MASK;
                  break;
        case 'r': /******** in file            ********/
                  fd = open(optarg, O_RDONLY);
                  break;
        case 's': /******** zero at end        ********/
                  flags |= MIFE_SLID;
                  break;
        case 't': /******** timeouts           ********/
                  tt = strtoul(optarg, NULL, 0) * 1000ULL;
                  break;
        case 'v': /******** verbose            ********/
                  BLIN_VERBOSE(flags);
                  break;
        case 'x': /******** end of line marker ********/
                  eoln = optarg;
                  break;
        default:
                  exit(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (argc & 1) exit(EX_USAGE);
    mf = mife_init(flags);
    if  (mife_ctl(mf, MIFE_CTLFDSC, fd)) {
        exit(EX_IOERR);
    }
    if  (tt && mife_ctl(mf, MIFE_CTLTIM1, tt)) {
        exit(EX_IOERR);
    }
    /****************************
     * [len off[ len off[...]]] *
     ****************************/
    c = 1;
    if  (argc == 0) {
        for (i = 0, of = 0; c > 0 && !(mf->flags & MIFE_EOFL); ++i, of += c) {
            if  (mife_ctl(mf, MIFE_CTLEOLI, eoln) < 0) err(EX_OSERR, "mife_eolinit");
            c = mife_ctl(mf, MIFE_CTLREAD, of, bl);
            if  (pr && c >= 0) {
                if  (!(pos = mife_point(mf, MIFE_POINOFF, of))) {
                    printf("i=%d %"BLIN_D" l=%"BLIN_D"->NO\n", i, c, bl);
                } else {
                    printf( "i=%d %"BLIN_D" l=%"BLIN_D"->%"BLIN_U" **"
                          , i, c, bl, (flags & MIFE_FULL) ? strlen((char*)pos) : c
                          );
                    for (j = 0; j < 8 && j < c; ++j) printf(" %02X", pos[j]);
                    printf("\n");
                    for (pl = 0; pl < c; ++pl) {
                        if  (!eoln && ((pos[pl] & 0xFF) != SEQN(of + pl))) {
                            printf( "off=%08"BLIN_O"X+%"BLIN_X" %02X<>%02X\n"
                                  , of, pl, pos[pl] & 0xFF, SEQN(of + pl)
                                  )
                            ;
        }   }   }   }   }
        printf("%d %"BLIN_D"\n", i, c);
    } else {
        for (i = 0; argc; ++i, argc -= 2) {
            bl = strtoul(*(argv++), NULL, 0);
            of = strtoq(*(argv++), NULL, 0);
            if  (mife_ctl(mf, MIFE_CTLEOLI, eoln) < 0) err(EX_OSERR, "mife_eolinit");
            c = mife_ctl(mf, MIFE_CTLREAD, of, bl);
            if  (pr && c >= 0) {
                if  (!(pos = mife_point(mf, MIFE_POINOFF, of))) {
                    printf("i=%d %"BLIN_D" l=%"BLIN_D"->NO\n", i, c, bl);
                } else {
                    printf( "i=%d %"BLIN_D" l=%"BLIN_D"->%"BLIN_U" **"
                          , i, c, bl, (flags & MIFE_FULL) ? strlen((char*)pos) : c
                          );
                    for (j = 0; j < 8 && j < c; ++j) printf(" %02X", pos[j]);
                    printf("\n");
                    for (pl = 0; pl < c; ++pl) {
                        if  (!eoln && ((pos[pl] & 0xFF) != SEQN(of + pl))) {
                            printf( "off=%08"BLIN_O"X+%"BLIN_X" %02X<>%02X\n"
                                  , of, pl, pos[pl] & 0xFF, SEQN(of + pl)
                                  )
                            ;
        }   }   }   }   }
        printf("%d %"BLIN_D"\n", i, c);
    }
    exit(EX_OK);
}
